<?php
declare(strict_types=1);

namespace FiloBlu\Refilo\Remote\Entity\Provider;

use FiloBlu\Refilo\Model\RelatedFactory;
use FiloBlu\Refilo\Model\UpsellFactory;
use FiloBlu\Refilo\Remote\Entity\EntityProviderInterface;
use FiloBlu\Refilo\Remote\Entity\ProductRelated;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Model\Product;
use Magento\Catalog\Model\ResourceModel\Product\Collection;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory;
use Magento\Framework\App\Area;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Store\Api\Data\StoreInterface;
use Magento\Store\Api\Data\WebsiteInterface;
use Magento\Store\Model\App\Emulation;
use Magento\Store\Model\StoreManagerInterface;
use Magento\TargetRule\Helper\Data;
use Magento\TargetRule\Model\Rule;
use Psr\Log\LoggerInterface;
use Throwable;

/**
 * Class RelatedProvider
 * @package FiloBlu\Refilo\Remote\Entity\Provider
 */
class RelatedProvider extends BaseProvider
{
    /**
     * @var UpsellFactory
     */
    private $upsellFactory;

    /**
     * @var CollectionFactory
     */
    private $productCollectionFactory;

    /**
     * @var RelatedFactory
     */
    private $relatedFactory;

    /**
     * @var Data
     */
    private $targetHelper;

    /**
     * @var Emulation
     */
    private $emulation;

    /**
     * @var LoggerInterface
     */
    private $logger;

    /**
     * @var StoreManagerInterface
     */
    private $storeManager;

    /**
     * RelatedProvider constructor.
     * @param CollectionFactory $productCollectionFactory
     * @param UpsellFactory $upsellFactory
     * @param RelatedFactory $relatedFactory
     * @param Emulation $emulation
     * @param Data $targetHelper
     * @param StoreManagerInterface $storeManager
     * @param ScopeConfigInterface $scopeConfig
     * @param LoggerInterface $logger
     */
    public function __construct(
        CollectionFactory $productCollectionFactory,
        UpsellFactory $upsellFactory,
        RelatedFactory $relatedFactory,
        Emulation $emulation,
        Data $targetHelper,
        StoreManagerInterface $storeManager,
        ScopeConfigInterface $scopeConfig,
        LoggerInterface $logger
    ) {
        parent::__construct($scopeConfig);
        $this->upsellFactory = $upsellFactory;
        $this->productCollectionFactory = $productCollectionFactory;
        $this->relatedFactory = $relatedFactory;
        $this->targetHelper = $targetHelper;
        $this->emulation = $emulation;
        $this->logger = $logger;
        $this->storeManager = $storeManager;
    }

    /**
     * @param StoreInterface $store
     * @return EntityProviderInterface
     */
    public function forStore(StoreInterface $store): EntityProviderInterface
    {
        return $this;
    }

    /**
     * @param WebsiteInterface $website
     * @return EntityProviderInterface
     */
    public function forWebsite(WebsiteInterface $website): EntityProviderInterface
    {
        return $this;
    }

    /**
     *
     */
    public function rewind()
    {
        $this->items = [];

        /** @var Collection $productCollection */
        $productCollection = $this->productCollectionFactory->create();

        if ($this->ids) {
            $productCollection->addFieldToFilter('entity_id', ['in' => $this->ids]);
        }

        $products = $productCollection->getItems();

        $defaultStore = $this->storeManager->getDefaultStoreView();

        $this->emulation->startEnvironmentEmulation(
            $defaultStore?$defaultStore->getId():0,
            Area::AREA_FRONTEND,
            true
        );

        try {
            $upsellBehaviour = (int)$this->targetHelper->getShowProducts(Rule::UP_SELLS);
            $relatedBehaviour = (int)$this->targetHelper->getShowProducts(Rule::RELATED_PRODUCTS);
            $crosssellBehaviour = (int)$this->targetHelper->getShowProducts(Rule::CROSS_SELLS);

            /** @var ProductInterface|Product $product */
            foreach ($products as $product) {
                $entityId = (int)$product->getId();

                $this->items[$entityId] = new ProductRelated([
                    'id'       => $entityId,
                    'upsell'   => $this->getUpSellProductIds($product, $upsellBehaviour),
                    'related'  => $this->getRelatedProductsIds($product, $relatedBehaviour),
                    'crossell' => $this->getCrossSellProductIds($product, $crosssellBehaviour)
                ]);
            }
        } catch (Throwable $throwable) {
            $this->logger->error($throwable->getMessage(), ['exception' => $throwable]);
        } finally {
            $this->emulation->stopEnvironmentEmulation();
        }

        if (!$this->hasReadHandler()) {
            return;
        }

        // TODO: Handle pagination
        // Temporary workaround for Morato
        $pageSize = $this->getBulkSize();
        $size = count($this->items);

        while ($size > $pageSize) {
            $data = array_slice($this->items, 0, $pageSize);
            $this->getReadHandler()->onRead($data, $pageSize, $this->getReadHandlerArguments());
            $size -= $pageSize;
        }

        if ($size > 0) {
            $this->getReadHandler()->onRead($this->items, $size, $this->getReadHandlerArguments());
        }
    }


    /**
     * @param Product $product
     * @param int $behaviour
     * @return array
     */
    public function getUpSellProductIds($product, $behaviour)
    {
        $ids = [];

        switch ($behaviour) {
            case Rule::BOTH_SELECTED_AND_RULE_BASED:
                $ids = array_merge(
                    $product->getUpSellProductIds(),
                    $this->upsellFactory->create()->setProduct($product)->getAllIds()
                );
                break;
            case Rule::SELECTED_ONLY:
                $ids = $product->getUpSellProductIds();
                break;
            case Rule::RULE_BASED_ONLY:
                $ids = $this->upsellFactory->create()->setProduct($product)->getAllIds();
                break;
        }

        return array_map('intval', $ids);
    }

    /**
     * @param Product $product
     * @param int $behaviour
     * @return array
     */
    public function getRelatedProductsIds($product, $behaviour)
    {
        $ids = [];

        switch ($behaviour) {
            case Rule::BOTH_SELECTED_AND_RULE_BASED:
                $ids = array_merge(
                    $product->getRelatedProductIds(),
                    $this->relatedFactory->create()->setProduct($product)->getAllIds()
                );
                break;
            case Rule::SELECTED_ONLY:
                $ids = $product->getRelatedProductIds();
                break;
            case Rule::RULE_BASED_ONLY:
                $ids = $this->relatedFactory->create()->setProduct($product)->getAllIds();
                break;
        }

        return array_map('intval', $ids);
    }


    /**
     * @param Product $product
     * @param $behaviour
     * @return array
     */
    public function getCrossSellProductIds($product, $behaviour)
    {
        $ids = [];

        switch ($behaviour) {
            case Rule::BOTH_SELECTED_AND_RULE_BASED:
                $ids = array_merge($product->getRelatedProductIds(), []);
                break;
            case Rule::SELECTED_ONLY:
                $ids = $product->getCrossSellProductIds();
                break;
            case Rule::RULE_BASED_ONLY:
                $ids = [];
                break;
        }

        return array_map('intval', $ids);
    }
}
